(***********************************************************************)
(*                                                                     *)
(*                        FoCaLize compiler                            *)
(*                                                                     *)
(*            Franois Pessaux                                         *)
(*            Pierre Weis                                              *)
(*            Damien Doligez                                           *)
(*                                                                     *)
(*                               LIP6  --  INRIA Rocquencourt          *)
(*                                                                     *)
(*  Copyright 2007, 2008 LIP6 and INRIA                                *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: parsetree_utils.ml,v 1.30 2008-12-01 12:45:40 weis Exp $ *)

let name_of_vname = function
  | Parsetree.Vlident s
  | Parsetree.Vuident s
  | Parsetree.Vpident s
  | Parsetree.Viident s
  | Parsetree.Vqident s -> s
;;



(* ************************************************************************* *)
(** {b Descr} : Module stuff to create sets of [Parsetree.vname]s with their
    type.
    This will serves to make sets of methods [vname]s in order to represent
    dependencies of methods of "Self" on methods of "Self".
    We keep the name's scheme in order to be able to annotate it if required
    during Coq code generation.

    {b Rem} : Not exported outside this module.                              *)
(* ************************************************************************* *)
module SelfDepMod = struct
  type t = (Parsetree.vname * Types.type_simple)
  let compare (n1, _) (n2, _) = compare n1 n2
end ;;
module SelfDepSet = Set.Make (SelfDepMod) ;;



(* ************************************************************************ *)
(* {b Descr} Allows to embedd in sets of dependencies on species parameters
   either the type scheme of the method if it is computational or the
   logical expression if it it a logical property. This comes from the fact
   that "type" of a computational method is its ML-like type but for
   logical methods (theorem/property), it is its logical property.

   {b Rem} : Exported outside this module.                                  *)
(* ************************************************************************ *)
type dependency_elem_type_kind =
  | DETK_computational of Types.type_simple
  | DETK_logical of Parsetree.logical_expr
;;


(* ************************************************************************* *)
(** {b Descr} : Module stuff to create sets of [Parsetree.vname]s with either
    their ML-like type or their logical expression.
    This will serves to make sets of methods [vname]s in order to represent
    dependencies of methods of "Self" on methods of species parameters.
    We keep the name's scheme or logical expression (depending on wether the
    method we depend on is computational or logical) in order to be able to
    annotate it or print its body (for logical expressions) if required
    during Coq code generation.

    {b Rem} : Not exported outside this module.                              *)
(* ************************************************************************* *)
module ParamDepMod = struct
  type t = (Parsetree.vname * dependency_elem_type_kind)
  let compare (n1, _) (n2, _) = compare n1 n2
end ;;
module ParamDepSet = Set.Make (ParamDepMod) ;;



let list_to_param_dep_set l =
  let rec rec_to_set accu = function
    | [] -> accu
    | h :: q ->
        rec_to_set (ParamDepSet.add h accu) q in
  rec_to_set ParamDepSet.empty l
;;



let param_dep_set_find predicate set =
  let found = ref None in
  try
    ParamDepSet.iter
      (fun elem ->
        if predicate elem then
          begin
          found := Some elem ;
          raise Exit
          end)
      set ;
      raise Not_found
  with Exit ->
    match !found with
     | None -> assert false
     | Some elem -> elem
;;



module VnameMod = struct
  type t = Parsetree.vname
  let compare = compare
end ;;
module VnameSet = Set.Make (VnameMod) ;;



(* ************************************************************************* *)
(* Parsetree.pattern -> Parsetree.vname list                                 *)
(** {b Descr} : Returns the list of local identifiers induced by a pattern.
        In other words, the list of identifiers a pattern binds and that
        in effect now are local idents in the expression using this pattern.

    {b Rem} : Exported outside this module.                                  *)
(* ************************************************************************* *)
let rec get_local_idents_from_pattern pat =
  match pat.Parsetree.ast_desc with
   | Parsetree.P_const _
   | Parsetree.P_wild -> []
   | Parsetree.P_var v -> [v]
   | Parsetree.P_as (p, v) -> v :: (get_local_idents_from_pattern p)
   | Parsetree.P_tuple ps
   | Parsetree.P_constr (_, ps) ->
       List.flatten (List.map get_local_idents_from_pattern ps)
   | Parsetree.P_record labs_pats ->
        List.flatten
         (List.map (fun (_, p) -> get_local_idents_from_pattern p) labs_pats)
   | Parsetree.P_paren p -> get_local_idents_from_pattern p
;;



let rec get_local_idents_and_types_from_pattern pat =
  match pat.Parsetree.ast_desc with
   | Parsetree.P_const _
   | Parsetree.P_wild -> []
   | Parsetree.P_var v -> [(v, pat.Parsetree.ast_type)]
   | Parsetree.P_as (p, v) ->
       (v, pat.Parsetree.ast_type) ::
       (get_local_idents_and_types_from_pattern p)
   | Parsetree.P_tuple ps
   | Parsetree.P_constr (_, ps) ->
       List.flatten (List.map get_local_idents_and_types_from_pattern ps)
   | Parsetree.P_record labs_pats ->
        List.flatten
         (List.map
	    (fun (_, p) -> get_local_idents_and_types_from_pattern p) labs_pats)
   | Parsetree.P_paren p -> get_local_idents_and_types_from_pattern p
;;



(* ********************************************************************* *)
(* string -> string                                                      *)
(** {b Descr} : Translate a FoCaLize operator name to a legal OCaml or Coq
              function name, preventing the versatile FoCaLize operators
              names from being lexically incorrect if straighforwardly
              converted into OCaml or Coq identifiers.
              The transformation is pretty simple and stupid (read
              sturdy :), replacing all the legal "symbolic" characters
              available for FoCaLize prefix/infix idents (extracted from
              the lexer definitions) by a regular string describing it.

    {b Rem} : Not exported outside this module.

    {b Rem} : This function should not be defined here but in the lexical
    analyser directory, in order to be up to date with the lexem
    definitions. To be even safer, we should define the full range of tokens
    in the [vname] type: we still lack the prefix upper and infix upper
    classes of idents; this would correct and simplify the definition of
    [parse_operator_string], which is still wrong, since it does not handle
    the bunch of particular cases it should treat to (re)-discover the true
    class of identifiers, if we want to correctly treat them.  *)

(* ********************************************************************* *)
let parse_operator_string op_string =
  let renamed_operator = ref "" in
  String.iter
    (fun character ->
      let str_tail =
        (match character with
         | '`' (* ` Helping emacs *) -> "_backquote_"
         | '~' -> "_tilda_"
         | '?' -> "_question_"
         | '$' -> "_dollar_"
         | '!' -> "_bang_"
         | '#' -> "_sharp_"
         | '+' -> "_plus_"
         | '-' -> "_dash_"
         | '*' -> "_star_"
         | '/' -> "_slash_"
         | '%' -> "_percent_"
         | '&' -> "_amper_"
         | '|' -> "_bar_"
         | ',' -> "_comma_"
         | ':' -> "_colon_"
         | ';' -> "_semi_"
         | '<' -> "_lt_"
         | '=' -> "_equal_"
         | '>' -> "_gt_"
         | '@' -> "_at_"
         | '^' -> "_hat_"
         | '\\' -> "_backslash_"
           (* In case we have a delimited ident with strange chars in it. *)
         | ' ' -> "_space_"
         | '.' -> "_dot_"
         | '(' -> "_lparen_"
         | ')' -> "_rparen_"
         | '[' -> "_lbracket_"
         | ']' -> "_rbracket_"
         | '{' -> "_lbrace_"
         | '}' -> "_rbrace_"
         | c ->
             (* For any other character, keep it unchanged. *)
             String.make 1 c) in
      (* Appending on string is not very efficient, but *)
      (* this should not be a real matter here ! *)
      renamed_operator := !renamed_operator ^ str_tail)
    op_string ;
  (* Just return the "translated" identifier name. *)
  !renamed_operator
;;

let vname_as_string_with_operators_expanded = function
  | Parsetree.Vlident s
  | Parsetree.Vqident s -> s
  | Parsetree.Vuident ("()" | "[]" | "::" as s) -> s
  | Parsetree.Vuident s ->
      assert (String.length s > 0);
      begin match s.[0] with
      | 'A' .. 'Z' -> s
      | _ -> parse_operator_string s end
  | Parsetree.Vpident s
  | Parsetree.Viident s -> parse_operator_string s
;;

(* ********************************************************************* *)
(* Format.formatter -> Parsetree.vname -> unit                           *)
(** {b Descr} : Pretty prints a [vname] value as an OCaml or Coq source ident.
    Because FoC allows more infix/prefix operators than OCaml or Coq
    syntax, it's impossible to crudely translate the string of the
    [vname] to OCaml or Coq.
    For instance, a FoC infix operator "( **+ )" has no equivalent in
    OCaml or Coq syntax : "( **+ )" is not a correct operator identifier
    according to OCaml or Coq.
    Then, instead of having particular cases for operators that can be
    straighforward translated (like "( +)") and the others, we adopt a
    uniform mapping for infix and prefix operators using the
    [vname_as_string_with_operators_expanded] function to transform
    infix/prefix operators names before printing and straighforwardly
    print other operators names.

    {b Rem} : Exported ouside this module.                               *)
(* ********************************************************************* *)
let pp_vname_with_operators_expanded ppf vname =
  Format.fprintf ppf "%s" (vname_as_string_with_operators_expanded vname)
;;

let type_coll_from_qualified_species (species_modname, species_vname) =
  (species_modname, (name_of_vname species_vname))
;;



(* ********************************************************************** *)
(** {b Descr} : Describes a species expression used a effective argument
    of a parametrised species. Since an effective parameter of a
    parametrised species can not have itself effective parameters, the
    only possible expressions are those denoting "Self" or another atomic
    species name.

    {b Rem}: Exported outside this module.                                  *)
(* ********************************************************************** *)
type simple_species_expr_as_effective_parameter =
  (** The name of the species used as species parameter is "Self". *)
  | SPE_Self
  (** The name of the species used as species parameter "IS" something else.
      We tag to know if it "IS" a species parameter, a toplevel collection
      or a toplevel species. *)
  | SPE_Species of (Parsetree.qualified_vname * Types.species_collection_kind)
  (** Regular expression used to instanciate a "in" (i.e. entity) parameter. *)
  | SPE_Expr_entity of Parsetree.expr
;;



type simple_species_expr = {
  sse_name : Parsetree.ident ;  (** Name of the base species. *)
  (** Effective arguments that are applied to it. *)
  sse_effective_args : simple_species_expr_as_effective_parameter list
  } ;;



(* ********************************************************************* *)
(* current_unit: Parsetree.module_name -> Parsetree.qualified_species -> *)
(*   Parsetree.ident                                                     *)
(** {b Descr} : Creates a fake [Parsetree.ident] from a
    [Parsetree.qualified_species] in order to be able to lookup in an
    environment. The ident is "fake" in the sense it has no location no
    documentation an no type. Then it must only b used when these fields
    are non-relevant.

    {b Rem} : Exported outside this module.                              *)
(* ********************************************************************* *)
let make_pseudo_species_ident ~current_unit (species_mod, species_name) =
  let ident_desc =
    if species_mod = current_unit then
      Parsetree.I_global (Parsetree.Vname species_name)
    else
      Parsetree.I_global (Parsetree.Qualified (species_mod, species_name)) in
  { Parsetree.ast_loc = Location.none ;
    Parsetree.ast_desc = ident_desc ;
    Parsetree.ast_doc = [] ;
    Parsetree.ast_type = Parsetree.ANTI_none }
;;


let make_concatenated_name_from_qualified_vname = function
  | Parsetree.Vname vname -> name_of_vname vname
  | Parsetree.Qualified (mod_name, vname) ->
      mod_name ^ "." ^(name_of_vname vname)
;;

let make_concatenated_name_with_operators_expanded_from_qualified_vname =
  function
  | Parsetree.Vname vname -> vname_as_string_with_operators_expanded vname
  | Parsetree.Qualified (mod_name, vname) ->
      mod_name ^ "." ^ (vname_as_string_with_operators_expanded vname)
;;
